#include "worker.h"
#include "pyremoter.h"
#include <regex>

class CPUStateReader : public Worker {
	public:

        /// @brief Get the latest fetched temperatures
        void getRecentTemperatures(std::vector<double>& a_vector){
            comm_mutex_.lock();
            a_vector = cpu_state_.cpuTemperatures_;
            comm_mutex_.unlock();
        }

		/// @brief Get the latest status of the job
		bool get_status()                             { return this->status_; }

		/// @brief Reset the Reader
		void reset() {
			this->status_ = true;
		}

        /// @brief Stop cycling
        void stop(){
            keep_running_.store(false);
        }

    private:

        boost::interprocess::interprocess_mutex comm_mutex_;

        std::atomic<bool> keep_running_;

        struct CPUState{
            CPUState():
                cpuTemperatures_({0.0, 0.0, 0.0, 0.0, 0.0})
            {
            }
            std::vector<double> cpuTemperatures_;
        };

        CPUState cpu_state_;

        std::unordered_map<std::string, int> cpu_temp_id_map_ = {{"Package id 0", 0}, {"Core 0", 1}, {"Core 1", 2}, {"Core 2", 3}, {"Core 3", 4}};

        std::string runCommand(std::string cmd)
        {
            char buffer[128];
            std::string result = "";
            FILE* pipe = popen(cmd.c_str(), "r");
            
            if (!pipe) throw std::runtime_error("popen() failed!");
            
            try {
                while (!feof(pipe)) {
                    if (fgets(buffer, 128, pipe) != NULL)
                    result += buffer;
                }
            } catch (...) {
                
            }
            
            pclose(pipe);

            return result;
        }

        /// fetch read cpu temperatures
		bool readCPUState(CPUState &a_cpu_state)
        {
            std::string sensors_installed_command_result = runCommand("dpkg -l | grep '^ii' | grep 'lm-sensors' | grep '1:3.4.0-4'");
            if (sensors_installed_command_result == ""){
                return false; // not installed
            }

            std::string sensors_command_result = runCommand("sensors");


            std::regex temperaturePattern(R"(([A-Za-z0-9' ']+):\s+([+\-]?[0-9.]+)\°C)");
            

            auto it = std::sregex_iterator(sensors_command_result.begin(), sensors_command_result.end(), temperaturePattern);
            auto end = std::sregex_iterator();


            boost::interprocess::scoped_lock lock(comm_mutex_);
            // Iterate over matches and extract sensor name and value
            for (; it != end; ++it) {
                std::smatch match = *it;
                std::string sensorName = match[1].str();
                double temperatureValue = std::stod(match[2].str());

                if (cpu_temp_id_map_.find(sensorName) != cpu_temp_id_map_.end()){
                    a_cpu_state.cpuTemperatures_[cpu_temp_id_map_[sensorName]] = temperatureValue;
                }
            }

            return true;
        }


		/// implement run for your specific Worker
		void run() {
            keep_running_.store(true);
            while(keep_running_.load()){
                this->status_ = readCPUState(cpu_state_);
                std::this_thread::sleep_for(std::chrono::milliseconds(500)); // 2 hz
            }
            this->finished();
		}

		bool status_  = true;
};